from typing import Union

from kivy.uix.button import Button
from kivy.uix.checkbox import CheckBox
from kivy.uix.label import Label
from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.spinner import Spinner
from kivy.uix.textinput import TextInput
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.widget import Widget


class LeftButton(Button):

    def __init__(self, index, **kwargs):
        super().__init__(**kwargs)
        self.pos_hint = {'x': .05 + .45 * index, 'y': 0}


class CoupledToggleButton(ToggleButton):

    def __init__(self, index, **kwargs):
        super().__init__(**kwargs)
        self.pos_hint = {'x': .05 + .45 * index, 'y': 0}


class CoupledRelativeLayout(RelativeLayout):

    def __init__(self, index, ratio=1, **kwargs):
        super().__init__(**kwargs)
        self.size_hint = (.575, .2 / ratio)
        self.pos_hint = {'x': 0., 'y': .95 - index * (self.size_hint[1] + .02)}


class WidgetLabelCoupledRelativeLayout(RelativeLayout):

    def __init__(self, widget: Widget, action, label: str, index: int, ratio: float = 1.0, **kwargs):
        super().__init__(**kwargs)
        self.size_hint = (.575, 1. / 6. / ratio)
        self.pos_hint = {'x': 0., 'y': .75 - index / 6. / ratio}
        self.add_widget(Label(text=label, size_hint=(.6, 1), pos_hint={'x': 0., 'y': 0.}))
        widget.size_hint = (.35, 1)
        widget.pos_hint = {'x': .6, 'y': 0.}
        if isinstance(widget, TextInput):
            widget.bind(text=action)
        elif isinstance(widget, CheckBox):
            widget.bind(active=action)
        else:
            raise NotImplementedError
        self.add_widget(widget)


class TextLabelCoupledRelativeLayout(WidgetLabelCoupledRelativeLayout):

    def __init__(self, label: str, text: str, filter: str, action, index: int, ratio: float = 1.0, **kwargs):
        super().__init__(TextInput(text=text, input_filter=filter), action, label, index, ratio, **kwargs)


class RadioLabelCoupledRelativeLayout(WidgetLabelCoupledRelativeLayout):

    def __init__(self, label: str, default: bool, action, index: int, ratio: float = 1.0, **kwargs):
        super().__init__(CheckBox(active=default), action, label, index, ratio, **kwargs)


class SpinnerLabelCoupledRelativeLayout(RelativeLayout):

    def __init__(self, label: str, default: str, options: list, action, index: int, ratio: float = 1.0, **kwargs):
        super().__init__(**kwargs)
        self.size_hint = (.575, 1. / 6. / ratio)
        self.pos_hint = {'x': 0., 'y': .75 - index / 6. / ratio}
        self.add_widget(Label(text=label, size_hint=(.6, 1), pos_hint={'x': 0., 'y': 0.}))
        spinner = Spinner(text=default, values=options, pos_hint={'center_x': .775, 'center_y': .5}, size_hint=(.35, 1))
        spinner.bind(text=action)
        self.add_widget(spinner)


class PanelBoxLayout(RelativeLayout):

    def __init__(self, controller, button_text, label_text, index=1, ratio=1,
                 spinner_text=None, spinner_list=None, param_function=None, **kwargs):
        super().__init__(**kwargs)
        self.controller = controller
        self.info_label = Label(text=label_text, size_hint=(.425, 1), pos_hint={'x': .575, 'y': 0})
        self.param_f = param_function
        self.spinner_text = spinner_text
        self.spinner_list = spinner_list
        self.ratio = ratio

        self.spinner_options = Spinner(
            pos_hint={'center_x': .275, 'center_y': .5}, size_hint=(.45, 1)
        )
        self.spinner_options.bind(text=self.select)
        self.go_button = LeftButton(1, text=button_text, disabled=True, on_press=self.go_action)

        self.main_panel = CoupledRelativeLayout(index, ratio)
        self.main_panel.add_widget(self.spinner_options)
        self.main_panel.add_widget(self.go_button)

    def select(self, spinner, text):
        pass

    def go_action(self, button):
        pass

    def enable(self):
        self.spinner_options.disabled = False

    def set_param(self, key, widget, value):
        self.param_f(key, value)

    def init(self):
        self.spinner_options.text = self.spinner_text
        task = self.controller.get_task_from_model()
        self.spinner_options.values = [k for k, v in self.spinner_list.items() if task in v[0]]
        self.go_button.disabled = True
        self.spinner_options.disabled = True
        self.parameter_panel.clear_widgets()
        self.set_info()

    def set_info(self):
        pass


class FeatureSelectionBoxLayout(RelativeLayout):

    def __init__(self, feature, role, ranking, action, plot_action, **kwargs):
        super().__init__(size_hint=(.24, 1. / 7.), **kwargs)
        self.buttons = {}
        self.label = Label(size_hint=(.7, 1.), pos_hint={'x': 0, 'y': 0},
                           halign='right', text_size=(self.width * 2, self.height * .2))
        self.set_text(feature, ranking)
        self.add_widget(self.label)
        for i, text in enumerate(['I', 'O']):
            button = ToggleButton(text=text, size_hint=(.1, .8), pos_hint={'x': .7 + .1 * i, 'y': .1},
                                  state='down' if role == text else 'normal', group=f'feature_{feature}')
            button.bind(on_press=action)
            self.add_widget(button)
            self.buttons[text] = button
        self.plot_button = ToggleButton(text='P', size_hint=(.1, .8), pos_hint={'x': .9, 'y': .1},
                                        state='down' if role == 'O' else 'normal')
        self.plot_button.bind(on_press=plot_action)
        self.add_widget(self.plot_button)

    def set_text(self, feature, ranking):
        text = feature if len(feature) < 16 else feature[:12] + '...'
        if ranking is not None:
            text += f' [{ranking:.2f}]'
        self.label.text = text


def create_param_layout(name: str, default: Union[str, bool, int, float], type, action, index: int, ratio: float):
    if isinstance(type, list):
        widget = SpinnerLabelCoupledRelativeLayout(name, default, type, action, index, ratio)
    elif type == 'bool':
        widget = RadioLabelCoupledRelativeLayout(f'{name}', default, action, index, ratio)
    else:
        widget = TextLabelCoupledRelativeLayout(f'{name} ({default})', '', type, action, index, ratio)
    return widget



from psyke.gui.model import TASKS
from matplotlib import colors, pyplot as plt

DATASET_MESSAGE = 'Select dataset'

INFO_DATASET_PREFIX = 'Dataset info:\n\n'

NO_DATASET = 'No dataset selected'

INFO_DATASET_MESSAGE = INFO_DATASET_PREFIX + NO_DATASET + '\n\n'

PREDICTOR_MESSAGE = 'Select predictor'

INFO_PREDICTOR_PREFIX = 'Predictor info:\n'

INFO_PREDICTOR_MESSAGE = f'\n\n{INFO_PREDICTOR_PREFIX}\nNo predictor trained\n\n\n\n\n\n'

PREDICTOR_PERFORMANCE_PREFIX = 'Predictor performance:\n'

EXTRACTOR_MESSAGE = 'Select extractor'

INFO_EXTRACTOR_PREFIX = 'Extractor info:\n'

INFO_EXTRACTOR_MESSAGE = f'\n\n{INFO_EXTRACTOR_PREFIX}\nNo extractor trained\n\n\n\n\n\n'

THEORY_PREFIX = 'Extracted theory'

THEORY_MESSAGE = 'No extractor trained\n\n\n'

EXTRACTOR_PERFORMANCE_PREFIX = 'Extractor performance:\n'

THEORY_ERROR_MESSAGE = {
    'amount': 'Too many extracted rules\n\n\n',
    'length': 'Too long clauses\n\n\n',
}

MARKERS = [
    ['x', '*', '+', 'o', 's', '^', 'D'],
    ['o', 's', '^', 'x', '*', '+', 'D'],
    ['s', '^', 'D', 'x', '*', '+', 'o']
]

COLORS = [
    ['tab:blue', 'tab:red', 'tab:green', 'tab:orange', 'tab:purple', 'tab:olive', 'tab:pink'],
    ['tab:orange', 'tab:purple', 'tab:olive', 'tab:blue', 'tab:red', 'tab:green', 'tab:pink'],
    ['tab:purple', 'tab:olive', 'tab:pink', 'tab:blue', 'tab:red', 'tab:green', 'tab:orange']
]

NAMES = {
    TASKS[0]: ['brg', 'opo', 'pop'],
    TASKS[1]: ['brg', 'rainbow', 'jet', 'turbo', 'cool', 'autumn', 'summer', 'spring', 'winter', 'plasma', 'viridis']
}

COLOR_MAPS = {
    TASKS[0]: [(name, colors.ListedColormap(cols)) for name, cols in zip(NAMES[TASKS[0]], COLORS)],
    TASKS[1]: [(name, plt.get_cmap(name)) for name in NAMES[TASKS[1]]]
}



from __future__ import annotations
import random as rnd
from itertools import product
from typing import Iterable
import numpy as np
import pandas as pd
from tuprolog.theory import Theory
from psyke import get_default_random_seed, PedagogicalExtractor
from psyke.utils import Target
from psyke.extraction.hypercubic import HyperCubeExtractor, Grid, HyperCube


class GridEx(PedagogicalExtractor, HyperCubeExtractor):
    """
    Explanator implementing GridEx algorithm, doi:10.1007/978-3-030-82017-6_2.
    """

    def __init__(self, predictor, grid: Grid, min_examples: int, threshold: float, normalization,
                 seed=get_default_random_seed()):
        super().__init__(predictor, normalization)
        self.grid = grid
        self.min_examples = min_examples
        self.threshold = threshold
        self.__generator = rnd.Random(seed)

    def _extract(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        if isinstance(np.array(self.predictor.predict(dataframe.iloc[0:1, :-1])).flatten()[0], str):
            self._output = Target.CLASSIFICATION
        surrounding = HyperCube.create_surrounding_cube(dataframe, output=self._output)
        surrounding.init_std(2 * self.threshold)
        self._iterate(surrounding, dataframe)
        return self._create_theory(dataframe, sort)

    def _ignore_dimensions(self) -> Iterable[str]:
        cube = self._hypercubes[0]
        return [d for d in cube.dimensions if all(c[d] == cube[d] for c in self._hypercubes)]

    def _iterate(self, surrounding: HyperCube, dataframe: pd.DataFrame):
        fake = dataframe.copy()
        prev = [surrounding]
        next_iteration = []

        for iteration in self.grid.iterate():
            next_iteration = []
            for cube in prev:
                to_split = []
                if cube.count(dataframe) == 0:
                    continue
                if cube.diversity < self.threshold:
                    self._hypercubes += [cube]
                    continue
                ranges = {}
                for (feature, (a, b)) in cube.dimensions.items():
                    bins = []
                    n_bins = self.grid.get(feature, iteration)
                    size = (b - a) / n_bins
                    for i in range(n_bins):
                        bins.append((a + size * i, a + size * (i + 1)))
                    ranges[feature] = bins
                for (pn, p) in enumerate(list(product(*ranges.values()))):
                    cube = self._default_cube()
                    for i, f in enumerate(dataframe.columns[:-1]):
                        cube.update_dimension(f, p[i])
                    n = cube.count(dataframe)
                    if n > 0:
                        fake = pd.concat([fake, cube.create_samples(self.min_examples - n, self.__generator)])
                        cube.update(fake, self.predictor)
                        to_split += [cube]
                to_split = self._merge(to_split, fake)
                next_iteration += [cube for cube in to_split]
            prev = next_iteration.copy()
        self._hypercubes += [cube for cube in next_iteration]

    @staticmethod
    def _find_couples(to_split: Iterable[HyperCube], not_in_cache: Iterable[HyperCube],
                      adjacent_cache: dict[tuple[HyperCube, HyperCube], str | None]) -> \
            Iterable[tuple[HyperCube, HyperCube, str]]:
        checked = []
        eligible = []
        for cube in to_split:
            checked.append(cube)
            for other_cube in [c for c in to_split if c not in checked]:
                if (cube in not_in_cache) or (other_cube in not_in_cache):
                    adjacent_cache[(cube, other_cube)] = cube.is_adjacent(other_cube)
                adjacent_feature = adjacent_cache[(cube, other_cube)]
                eligible.append((cube, other_cube, adjacent_feature))
        return [couple for couple in eligible if couple[2] is not None]

    def _evaluate_merge(self, not_in_cache: Iterable[HyperCube],
                        dataframe: pd.DataFrame, feature: str,
                        cube: HyperCube, other_cube: HyperCube,
                        merge_cache: dict[(HyperCube, HyperCube), HyperCube | None]) -> bool:
        if (cube in not_in_cache) or (other_cube in not_in_cache):
            merged_cube = cube.merge_along_dimension(other_cube, feature)
            merged_cube.update(dataframe, self.predictor)
            merge_cache[(cube, other_cube)] = merged_cube
        return cube.output == other_cube.output if self._output == Target.CLASSIFICATION else \
            merge_cache[(cube, other_cube)].diversity < self.threshold

    def _merge(self, to_split: Iterable[HyperCube], dataframe: pd.DataFrame) -> Iterable[HyperCube]:
        not_in_cache = [cube for cube in to_split]
        adjacent_cache = {}
        merge_cache = {}
        # TODO: refactor this. A while true with a break is as ugly as hunger.
        while True:
            to_merge = [([cube, other_cube], merge_cache[(cube, other_cube)]) for cube, other_cube, feature in
                        GridEx._find_couples(to_split, not_in_cache, adjacent_cache) if
                        self._evaluate_merge(not_in_cache, dataframe, feature, cube, other_cube, merge_cache)]
            if len(to_merge) == 0:
                break
            sorted(to_merge, key=lambda c: c[1].diversity)
            best = to_merge[0]
            to_split = [cube for cube in to_split if cube not in best[0]] + [best[1]]
            not_in_cache = [best[1]]
        return to_split



import numpy as np
from psyke.extraction.trepan.utils import Node, Split, SplitLogic
from psyke import DiscreteFeature, PedagogicalExtractor
from psyke.utils.logic import create_term, create_variable_list, create_head
from psyke.utils.sorted import SortedList
from tuprolog.core import Var, Struct, clause
from tuprolog.theory import MutableTheory, mutable_theory, Theory
from typing import Iterable, Union, Any
import pandas as pd


class Trepan(PedagogicalExtractor):

    def __init__(self, predictor, discretization: Iterable[DiscreteFeature], min_examples: int = 0, max_depth: int = 3,
                 split_logic: SplitLogic = SplitLogic.DEFAULT):
        super().__init__(predictor, discretization)
        self.min_examples = min_examples
        self.max_depth = max_depth
        self.split_logic = split_logic
        self._root: Node

    @property
    def n_rules(self):
        return sum(1 for _ in self._root)

    def _best_split(self, node: Node, names: Iterable[str]) -> Union[tuple[Node, Node], None]:
        if node.samples.shape[0] < self.min_examples:
            raise NotImplementedError()
        if node.n_classes == 1:
            return None
        splits = Trepan._create_splits(node, names)
        return None if len(splits) == 0 or splits[0].children[0].depth > self.max_depth else splits[0].children

    def _compact(self):
        nodes = [self._root]
        while len(nodes) > 0:
            node = nodes.pop()
            for item in self._nodes_to_remove(node, nodes):
                node.children.remove(item)
                node.children += item.children

    def _create_body(self, variables: dict[str, Var], node: Node) -> Iterable[Struct]:
        result = []
        for constraint, value in node.constraints:
            feature: DiscreteFeature = [d for d in self.discretization if constraint in d.admissible_values][0]
            result.append(create_term(variables[feature.name], feature.admissible_values[constraint], value == 1.0))
        return result

    @staticmethod
    def _create_samples(node: Node, column: str, value: float) -> pd.DataFrame:
        return node.samples.loc[node.samples[column] == value]

    @staticmethod
    def _create_split(node: Node, column: str) -> Union[Split, None]:
        true_examples = Trepan._create_samples(node, column, 1.0)
        false_examples = Trepan._create_samples(node, column, 0.0)
        true_constrains = list(node.constraints) + [(column, 1.0)]
        false_constrains = list(node.constraints) + [(column, 0.0)]
        true_node = Node(true_examples, node.n_examples, true_constrains, depth=node.depth + 1)\
            if true_examples.shape[0] > 0 else None
        false_node = Node(false_examples, node.n_examples, false_constrains, depth=node.depth + 1)\
            if false_examples.shape[0] > 0 else None
        return None if true_node is None or false_node is None else Split(node, (true_node, false_node))

    @staticmethod
    def _create_splits(node: Node, names: Iterable[str]) -> SortedList[Split]:
        splits, constrains = Trepan._init_splits(node)
        for column in names:
            if column not in constrains:
                split = Trepan._create_split(node, column)
                if split is not None:
                    splits.add(split)
        return splits

    def _create_theory(self, name: str, sort: bool = True) -> MutableTheory:
        theory = mutable_theory()
        for node in self._root:
            variables = create_variable_list(self.discretization, sort=sort)
            theory.assertZ(
                clause(
                    create_head(name, list(variables.values()), str(node.dominant)),
                    self._create_body(variables, node)
                )
            )
        return theory

    def _init(self, dateset: pd.DataFrame) -> SortedList[Node]:
        self._root = Node(dateset, dateset.shape[0])
        queue: SortedList[Node] = SortedList(lambda x, y: int(x.priority - y.priority))
        queue.add(self._root)
        return queue

    @staticmethod
    def _init_splits(node: Node) -> tuple[SortedList[Split], Iterable[str]]:
        return SortedList(lambda x, y: int(x.priority - y.priority)),\
               set(constraint[0] for constraint in node.constraints)

    @staticmethod
    def _nodes_to_remove(node: Node, nodes: list[Node]) -> list[Node]:
        to_remove = []
        for child in node.children:
            if node.dominant == child.dominant and len(child.children) == 1:
                to_remove.append(child)
                nodes.append(node)
            else:
                nodes.append(child)
        return to_remove

    @staticmethod
    def _internal_predict(x: pd.Series, node: Node, categories: Iterable) -> Any:
        for child in node.children:
            skip = False
            for constraint, value in child.constraints:
                if x[constraint] != value:
                    skip = True
                    continue
            if not skip:
                return Trepan._internal_predict(x, child, categories)
        return node.dominant  # Alternatively node.dominant index in categories

    def _optimize(self) -> None:
        n, nodes = 0, [self._root]
        while len(nodes) > 0:
            n += Trepan._remove_nodes(nodes)
        self._compact() if n == 0 else self._optimize()

    @staticmethod
    def _remove_nodes(nodes: list[Node]) -> int:
        node = nodes.pop()
        to_remove = [child for child in node.children if len(child.children) == 0 and node.dominant == child.dominant]
        for child in to_remove:
            node.children.remove(child)
        for child in node.children:
            if len(child.children) > 0:
                nodes.append(child)
        return len(to_remove)

    def _extract(self, dataframe: pd.DataFrame, mapping: dict[str: int] = None, sort: bool = True) -> Theory:
        queue = self._init(dataframe)
        while len(queue) > 0:
            node = queue.pop()
            if self.split_logic == SplitLogic.DEFAULT:
                best: Union[tuple[Node, Node], None] = self._best_split(node, dataframe.columns[:-1])
                if best is None:
                    continue
            else:
                raise Exception('Illegal split logic')
            queue.add_all(best)
            node.children += list(best)
        self._optimize()
        return self._create_theory(dataframe.columns[-1], sort)

    def _predict(self, dataframe: pd.DataFrame) -> Iterable:
        return np.array(
            [Trepan._internal_predict(sample, self._root, dataframe.columns[-1]) for _, sample in dataframe.iterrows()]
        )



from enum import Enum
import numpy as np
import pandas as pd

from psyke.utils import Target


class Objective(Enum):
    MODEL = 1,
    DATA = 2


class Optimizer:
    def __init__(self, predictor, algorithm, dataframe: pd.DataFrame, max_mae_increase: float = 1.2,
                 min_rule_decrease: float = 0.9, readability_tradeoff: float = 0.1, max_depth: int = 10,
                 patience: int = 5, objective: Objective = Objective.MODEL, normalization=None):
        self.predictor = predictor
        self.algorithm = algorithm
        self.dataframe = dataframe
        self.max_mae_increase = max_mae_increase
        self.min_rule_decrease = min_rule_decrease
        self.readability_tradeoff = readability_tradeoff
        self.patience = patience
        self.max_depth = max_depth
        self.objective = objective
        self.model_mae = abs(self.predictor.predict(dataframe.iloc[:, :-1]).flatten() -
                             self.dataframe.iloc[:, -1].values).mean()
        self.params = None
        self.normalization = normalization

    def _depth_improvement(self, first, second):
        if second[0] == first[0]:
            return (first[1] - second[1]) * 2
        return 1 / (
                (1 - second[0] / first[0]) ** self.readability_tradeoff *
                np.ceil(second[1] / self.readability_tradeoff) / np.ceil(first[1] / self.readability_tradeoff)
        )

    @staticmethod
    def _best(params):
        param_dict = {Optimizer.__score(t): t for t in params}
        min_param = min(param_dict)
        return min_param, param_dict[min_param]

    @staticmethod
    def __score(param):
        return param[0] * np.ceil(param[1] / 5)

    def _best_param(self, param):
        param_dict = {t[param]: t for t in self.params}
        min_param = min(param_dict)
        return min_param, param_dict[min_param]

    def get_best(self):
        names = [self.algorithm, "  MAE  ", "N rules"]
        params = [Optimizer._best(self.params), self._best_param(0), self._best_param(1)]
        for n, p in zip(names, params):
            self._print_params(n, p[1])
            print()
        return Optimizer._best(self.params)[1], self._best_param(0)[1], self._best_param(1)[1]

    def _print_params(self, n, param):
        raise NotImplementedError



from kivy.uix.screenmanager import ScreenManager

from psyke.gui.model import DatasetError, PredictorError, SVMError
from psyke.gui.view.View import MainScreen


class Controller:

    def __init__(self, model):
        self.model = model

        self.screen_manager = ScreenManager()
        self.main_screen = MainScreen(name="main")
        self.screen_manager.add_widget(self.main_screen)

        self.view = self.main_screen
        self.view.init(self)

    @property
    def screen(self):
        return self.screen_manager

    def get_task_from_model(self):
        return self.model.task

    def get_dataset_from_model(self):
        return self.model.dataset

    def get_data_from_model(self):
        return self.model.data, self.model.pruned_data, self.model.preprocessing_action, self.model.preprocessing

    def get_data_rankings_from_model(self):
        return self.model.ranked_data

    def get_predictor_from_model(self):
        return self.model.predictor_name, self.model.predictor, self.model.predictor_params

    def get_extractor_from_model(self):
        return self.model.extractor_name, self.model.extractor, self.model.extractor_params

    def get_test_set_from_model(self):
        return self.model.test, self.model.preprocessing_action, self.model.preprocessing

    def get_theory_from_model(self):
        return self.model.theory

    def select_task(self, task):
        self.model.reset_dataset()
        self.model.reset_preprocessing()
        self.model.select_task(task)
        self.reset_dataset()

    def select_preprocessing(self, action, value):
        if value:
            self.model.select_preprocessing(action)
        else:
            self.model.reset_preprocessing()
        self.load_dataset()

    def select_colormap(self, cmap):
        self.model.select_colormap(cmap)

    def select_dataset(self, dataset):
        self.model.reset_dataset()
        self.model.select_dataset(dataset)
        self.reset_predictor()
        self.view.data_panel.disable()
        self.view.data_panel.set_info()
        self.view.feature_panel.set_info()
        self.view.plot_panel.clear_data()

    def select_predictor(self, predictor):
        self.model.reset_predictor()
        self.model.select_predictor(predictor)
        self.reset_extractor()
        self.view.predictor_panel.set_info()
        self.view.plot_panel.clear_predictor()

    def select_extractor(self, extractor):
        self.model.reset_extractor()
        self.model.select_extractor(extractor)
        self.view.extractor_panel.set_info()
        self.view.theory_panel.set_info()
        self.view.plot_panel.clear_extractor()

    def reload_dataset(self, features):
        ret = True
        try:
            self.model.select_features(features)
        except ValueError as e:
            self.view.feature_panel.set_alert('Cannot calculate ranking for these features')
            ret = False
        self.reset_predictor()
        self.view.data_panel.set_info()
        self.view.predictor_panel.enable()
        self.view.plot_panel.clear_data()
        return ret

    def plot(self, features, plot_features, save=False):
        inputs = [k for k, v in features.items() if v == 'I' and k in plot_features]
        output = [k for k, v in features.items() if v == 'O' and k in plot_features][0]
        self.model.plot(inputs, output, save)
        self.view.plot_panel.set_info()

    def get_plots_from_model(self):
        return self.model.data_plot, self.model.predictor_plot, self.model.extractor_plot

    def reset_dataset(self):
        self.model.reset_dataset()
        self.reset_predictor()
        self.view.data_panel.init()
        self.view.feature_panel.init()
        self.view.data_panel.set_info()
        self.view.feature_panel.set_info()
        self.view.plot_panel.clear_data()

    def reset_predictor(self):
        self.model.reset_predictor()
        self.reset_extractor()
        self.view.predictor_panel.init()
        self.view.predictor_panel.set_info()
        self.view.plot_panel.clear_predictor()

    def reset_extractor(self):
        self.model.reset_extractor()
        self.view.extractor_panel.init()
        self.view.theory_panel.init()
        self.view.extractor_panel.set_info()
        self.view.theory_panel.set_info()
        self.view.plot_panel.clear_extractor()

    def load_dataset(self):
        self.model.reset_dataset(True)
        self.reset_predictor()
        self.view.feature_panel.init()
        try:
            self.model.load_dataset()
            self.view.data_panel.enable()
            self.view.predictor_panel.enable()
            self.view.feature_panel.set_info()
        except DatasetError as e:
            self.view.feature_panel.set_alert(e.message)
        self.view.data_panel.set_info()
        self.view.plot_panel.clear_data()

    def train_predictor(self):
        self.reset_extractor()
        self.view.plot_panel.clear_predictor()
        try:
            self.model.train_predictor()
            self.view.extractor_panel.enable()
            self.view.predictor_panel.set_info()
            self.view.feature_panel.reset_alert()
        except (PredictorError, SVMError) as e:
            self.view.feature_panel.set_alert(e.message)

    def train_extractor(self):
        self.view.plot_panel.clear_extractor()
        self.model.train_extractor()
        self.view.extractor_panel.set_info()
        self.view.theory_panel.set_info()

    def set_predictor_param(self, key, value):
        self.model.set_predictor_param(key, value)

    def set_extractor_param(self, key, value):
        self.model.set_extractor_param(key, value)



from kivy.garden.matplotlib.backend_kivyagg import FigureCanvasKivyAgg
from kivy.uix.label import Label
from kivy.uix.relativelayout import RelativeLayout


class PlotPanel(RelativeLayout):

    def __init__(self, controller, **kwargs):
        super().__init__(**kwargs)
        self.controller = controller

        self.data_plot_container = RelativeLayout(pos_hint={'x': 0., 'y': 0.}, size_hint=(.33, .965))
        self.predictor_plot_container = RelativeLayout(pos_hint={'x': .33, 'y': 0.},  size_hint=(.33, .965))
        self.extractor_plot_container = RelativeLayout(pos_hint={'x': .66, 'y': 0.},  size_hint=(.33, .965))

        self.data_plot = None
        self.predictor_plot = None
        self.extractor_plot = None

        self.add_widget(self.data_plot_container)
        self.add_widget(self.predictor_plot_container)
        self.add_widget(self.extractor_plot_container)

        self.init()

    def init(self):
        self.clear_data()

    def __add_widgets(self):
        self.data_plot_container.clear_widgets()
        self.data_plot_container.add_widget(self.data_plot)
        self.predictor_plot_container.clear_widgets()
        self.predictor_plot_container.add_widget(self.predictor_plot)
        self.extractor_plot_container.clear_widgets()
        self.extractor_plot_container.add_widget(self.extractor_plot)

    def set_info(self):
        plots = self.controller.get_plots_from_model()
        if plots[0] is not None:
            self.data_plot = FigureCanvasKivyAgg(plots[0])
        if plots[1] is not None:
            self.predictor_plot = FigureCanvasKivyAgg(plots[1])
        if plots[2] is not None:
            self.extractor_plot = FigureCanvasKivyAgg(plots[2])
        self.__add_widgets()

    def clear_data(self):
        self.data_plot = Label(text='Load a dataset\n\n', color=(0, 0, 0))
        self.clear_predictor()

    def clear_predictor(self):
        self.predictor_plot = Label(text='Train a predictor\n\n', color=(0, 0, 0))
        self.clear_extractor()

    def clear_extractor(self):
        self.extractor_plot = Label(text='Train an extractor\n\n', color=(0, 0, 0))
        self.__add_widgets()



from typing import Callable

import numpy as np
from parameterized import parameterized_class
from sklearn.model_selection import train_test_split
from tuprolog.solve.prolog import prolog_solver
from tuprolog.theory import mutable_theory

from psyke import Extractor
from psyke.utils import get_default_precision
from psyke.utils.logic import data_to_struct, get_in_rule, get_not_in_rule
from test import get_dataset, get_model
import unittest


# TODO: should be refactored using the a .csv file
@parameterized_class([{"dataset": "iris", "predictor": "DTC", "task": "extraction"},
                      {"dataset": "house", "predictor": "DTR", "task": "hypercubic"}])
class TestSimplifiedCart(unittest.TestCase):

    def test_equality(self):
        dataset = get_dataset(self.dataset)
        dataset = dataset.reindex(sorted(dataset.columns[:-1]) + [dataset.columns[-1]], axis=1)
        train, test = train_test_split(dataset, test_size=0.5)
        tree = get_model(self.predictor, {})
        tree.fit(train.iloc[:, :-1], train.iloc[:, -1])
        extractor = Extractor.cart(tree, simplify=False)
        theory = extractor.extract(train)
        simplified_extractor = Extractor.cart(tree)
        simplified_theory = simplified_extractor.extract(train)

        index = test.shape[1] - 1
        is_classification = isinstance(test.iloc[0, -1], str)
        cast: Callable = lambda x: (str(x) if is_classification else float(x.value))

        solver = prolog_solver(static_kb=mutable_theory(theory).assertZ(get_in_rule()).assertZ(get_not_in_rule()))
        substitutions = [solver.solveOnce(data_to_struct(data)) for _, data in test.iterrows()]
        expected = [cast(query.solved_query.get_arg_at(index)) for query in substitutions]

        simplified_solver = \
            prolog_solver(static_kb=mutable_theory(simplified_theory).assertZ(get_in_rule()).assertZ(get_not_in_rule()))
        simplified_substitutions = [simplified_solver.solveOnce(data_to_struct(data)) for _, data in test.iterrows()]
        simplified_expected = [cast(query.solved_query.get_arg_at(index)) for query in simplified_substitutions]

        if isinstance(test.iloc[0, -1], str):
            self.assertTrue(all(np.array(extractor.predict(test.iloc[:, :-1])) ==
                                np.array(simplified_extractor.predict(test.iloc[:, :-1]))))
            self.assertEqual(expected, simplified_expected)
        else:
            self.assertTrue(max(abs(np.array(extractor.predict(test.iloc[:, :-1])) -
                                    np.array(simplified_extractor.predict(test.iloc[:, :-1])))
                                ) < get_default_precision())
            self.assertTrue(max(abs(np.array(expected) - np.array(simplified_expected))) < get_default_precision())


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pandas as pd
from enum import Enum
from psyke import Extractor
from psyke.extraction.hypercubic import Grid, FeatureRanker
from psyke.extraction.hypercubic.strategy import AdaptiveStrategy, FixedStrategy
from psyke.tuning import Objective, Optimizer


class PEDRO(Optimizer):
    class Algorithm(Enum):
        GRIDEX = 1,
        GRIDREX = 2

    def __init__(self, predictor, dataframe: pd.DataFrame, max_mae_increase: float = 1.2,
                 min_rule_decrease: float = 0.9, readability_tradeoff: float = 0.1, max_depth: int = 3,
                 patience: int = 3, algorithm: Algorithm = Algorithm.GRIDREX, objective: Objective = Objective.MODEL,
                 normalization=None):
        super().__init__(predictor, algorithm, dataframe, max_mae_increase, min_rule_decrease, readability_tradeoff,
                         max_depth, patience, objective, normalization)
        self.ranked = FeatureRanker(dataframe.columns[:-1]).fit(predictor, dataframe.iloc[:, :-1]).rankings()

    def __search_threshold(self, grid, critical, max_partitions):
        step = self.model_mae / 2.0
        threshold = self.model_mae * 0.5
        params = []
        patience = self.patience
        while patience > 0:
            print("{}. {}. Threshold = {:.2f}. ".format(self.algorithm, grid, threshold), end="")
            extractor = Extractor.gridrex(self.predictor, grid, threshold=threshold, normalization=self.normalization) \
                if self.algorithm == PEDRO.Algorithm.GRIDREX \
                else Extractor.gridex(self.predictor, grid, threshold=threshold, normalization=self.normalization)
            _ = extractor.extract(self.dataframe)
            mae, n = (extractor.mae(self.dataframe, self.predictor) if self.objective == Objective.MODEL else
                      extractor.mae(self.dataframe)), extractor.n_rules
            print("MAE = {:.2f}, {} rules".format(mae, n))

            if len(params) == 0:
                params.append((mae, n, threshold, grid))
                threshold += step
                continue

            if n > max_partitions:
                break

            if n == 1:
                params.append((mae, n, threshold, grid))
                break

            if mae > params[0][0] * self.max_mae_increase:
                break

            improvement = (params[-1][0] / mae) + (1 - n / params[-1][1])

            if improvement <= 1 or n > np.ceil(params[-1][1] * self.min_rule_decrease):
                patience -= 1
                step = max(step, abs(mae - threshold) / max(patience, 1))
            elif not critical:
                patience = self.patience
            if mae != params[-1][0] or n != params[-1][1]:
                params.append((mae, n, threshold, grid))
            threshold += step
        return params

    def __search_depth(self, strategy, critical, max_partitions):
        params = []
        best = None

        for iterations in range(self.max_depth):
            grid = Grid(iterations + 1, strategy)
            p = self.__search_threshold(grid, critical, max_partitions)
            b = Optimizer._best(p)[1]
            print()
            improvement = self._depth_improvement(
                [best[0], best[1]], [b[0], b[1]]
            ) if best is not None else np.inf

            best = b
            params += p

            if len(params) > 1 and improvement < 1.2:
                break
        return params

    def __contains(self, strategies, strategy):
        for s in strategies:
            if strategy.equals(s, self.dataframe.columns[:-1]):
                return True
        return False

    def search(self):
        base_strategy = FixedStrategy(2)
        strategies = [base_strategy, FixedStrategy(3)]

        base_partitions = base_strategy.partition_number(self.dataframe.columns[:-1])

        for n in [2, 3, 5, 10]:
            for th in [0.99, 0.75, 0.67, 0.5, 0.3]:
                strategy = AdaptiveStrategy(self.ranked, [(th, n)])
                if strategy.partition_number(self.dataframe.columns[:-1]) < base_partitions * 3 and \
                        not self.__contains(strategies, strategy):
                    strategies.append(strategy)

        for (a, b) in [(0.33, 0.67), (0.25, 0.75), (0.1, 0.9)]:
            strategy = AdaptiveStrategy(self.ranked, [(a, 2), (b, 3)])
            if strategy.partition_number(self.dataframe.columns[:-1]) < base_partitions * 3 and \
                    not self.__contains(strategies, strategy):
                strategies.append(strategy)

        avg = 0.
        for strategy in strategies:
            avg += strategy.partition_number(self.dataframe.columns[:-1])
        avg /= len(strategies)

        params = []
        for strategy in strategies:
            params += self.__search_depth(strategy,
                                          strategy.partition_number(self.dataframe.columns[:-1]) > avg,
                                          base_partitions * 3)
        self.params = params

    def _print_params(self, name, params):
        print("**********************")
        print(f"Best {name}")
        print("**********************")
        print(f"MAE = {params[0]:.2f}, {params[1]} rules")
        print(f"Threshold = {params[2]:.2f}")
        print(f"Iterations = {params[3].iterations}")
        print(f"Strategy = {params[3].strategy}")



import unittest
from tuprolog.theory import mutable_theory, theory
from tuprolog.theory.parsing import parse_theory
from psyke.utils.logic import prune


class TestPrune(unittest.TestCase):

    def test_prune_documentation(self):
        theory1 = "c(A, B, C, D, positive) :- ('=<'(A, 1), '>'(B, 2), '=='(C, 0)). " \
                  + "c(A, B, C, D, positive) :- ('=<'(A, 1), '>'(B, 2))."
        pruned1 = "c(A, B, C, D, positive) :- ('=<'(A, 1), '>'(B, 2))."

        theory2 = "c(A, B, C, D, positive) :- ('=<'(A, 1), '>'(B, 2), '=='(C, 0)). " \
                  + "c(A, B, C, D, positive) :- ('=<'(A, 1.3), '>'(B, 1.8))."
        pruned2 = "c(A, B, C, D, positive) :- ('=<'(A, 1.3), '>'(B, 1.8))."

        theory3 = "c(A, B, C, D, positive) :- ('=<'(A, 1.3), '>'(B, 1.8)). " \
                  + "c(A, B, C, D, positive) :- ('=<'(A, 1), '>'(B, 2), '=='(C, 0))."
        pruned3 = pruned2

        self.assertTrue(theory(parse_theory(pruned1)).equals(prune(mutable_theory(parse_theory(theory1))), False))
        self.assertTrue(theory(parse_theory(pruned2)).equals(prune(mutable_theory(parse_theory(theory2))), False))
        self.assertTrue(theory(parse_theory(pruned3)).equals(prune(mutable_theory(parse_theory(theory3))), False))

    def test_prune_success(self):
        textual_theory = "p(X, Y, inside) :- ('=<'(X, 1), '>'(Y, 2)). " \
                         + "p(X, Y, inside) :- ('=<'(X, 0.5), '>'(Y, 3))."
        textual_pruned_theory = "p(X, Y, inside) :- ('=<'(X, 1), '>'(Y, 2))."
        long_theory = mutable_theory(parse_theory(textual_theory))
        pruned_theory = theory(parse_theory(textual_pruned_theory))

        self.assertTrue(pruned_theory.equals(prune(long_theory), False))

    def test_prune_not_applied(self):
        textual_theory = "p(PL, PW, SL, SW, versicolor) :-  '=<'(SW, 3.6). " \
                         + "p(PL, PW, SL, SW, versicolor) :- ('=<'(PW, 0.35), '=<'(SL, 5.35), '=<'(SW, 3.9))."
        textual_pruned_theory = textual_theory
        long_theory = mutable_theory(parse_theory(textual_theory))
        pruned_theory = theory(parse_theory(textual_pruned_theory))

        self.assertTrue(pruned_theory.equals(prune(long_theory), False))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from psyke import DiscreteFeature
from typing import Iterable
import pandas as pd


class Rule:

    def __init__(self, true_predicates: list[str], false_predicates: list[str]):
        self.true_predicates = true_predicates
        self.false_predicates = false_predicates

    def __contains__(self, other: Rule) -> bool:
        return all([predicate in other.true_predicates for predicate in self.true_predicates]) and\
               all([predicate in other.false_predicates for predicate in self.false_predicates])

    def __eq__(self, other: Rule) -> bool:
        return self.true_predicates == other.true_predicates and self.false_predicates == other.false_predicates

    def __hash__(self) -> int:
        return hash(self.true_predicates) + hash(self.false_predicates)

    def reduce(self, features: Iterable[DiscreteFeature]) -> Rule:
        to_be_removed = [item for tp in self.true_predicates
                         for feature in features if tp in feature.admissible_values
                         for item in feature.admissible_values.keys()]
        return Rule(self.true_predicates, [fp for fp in self.false_predicates if fp not in to_be_removed])

    def to_lists(self) -> list[list[str]]:
        return [self.true_predicates.copy(), self.false_predicates.copy()]


class IndexedRuleSet(dict[int, list[Rule]]):

    def flatten(self) -> list[tuple[int, Rule]]:
        return [(key, value) for key, values in self.items() for value in values]

    def optimize(self) -> IndexedRuleSet:
        useless_rules = [item for key, entry in self.items() for item in IndexedRuleSet._useless_rules(key, entry)]
        for rule in useless_rules:
            self[rule[0]].remove(rule[1])
        return self

    @staticmethod
    def _useless_rules(key, rules: list[Rule]) -> list[(int, Rule)]:
        return [
            (key, rule) for rule in rules
            if any(rule in other_rule for other_rule in rules if other_rule != rule)
        ]

    @staticmethod
    def create_indexed_ruleset(dataset: pd.DataFrame) -> IndexedRuleSet:
        return IndexedRuleSet({index: [] for index, _ in enumerate(set(dataset.iloc[:, -1]))})


from __future__ import annotations
from onnxconverter_common import DataType, FloatTensorType, Int64TensorType, StringTensorType
from skl2onnx import convert_sklearn
from sklearn.model_selection import train_test_split
from tensorflow.keras import Model
from tensorflow.python.saved_model.save import save
from tuprolog.solve.prolog import prolog_solver
from psyke.extraction.hypercubic import Grid, FeatureRanker
from psyke.utils.dataframe import get_discrete_dataset
from psyke.utils.logic import data_to_struct, get_in_rule, get_not_in_rule
from psyke.extraction.hypercubic.strategy import AdaptiveStrategy, FixedStrategy
from test import get_dataset, get_extractor, get_schema, get_model
from test.resources.predictors import get_predictor_path
from test.resources.tests import test_cases
from tuprolog.theory import Theory, mutable_theory
from tuprolog.theory.parsing import parse_theory
from typing import Iterable, Callable
import ast
import numpy as np
import onnxruntime as rt
import os
import pandas as pd
from psyke import get_default_random_seed


def initialize(file: str) -> list[dict[str:Theory]]:
    for row in test_cases(file):
        params = dict() if row['extractor_params'] == '' else ast.literal_eval(row['extractor_params'])
        dataset = get_dataset(row['dataset'])

        # Dataset's columns are sorted due to alphabetically sorted extracted rules.
        # columns = sorted(dataset.columns[:-1]) + [dataset.columns[-1]]
        # dataset = dataset.reindex(columns, axis=1)

        training_set, test_set = train_test_split(dataset, test_size=0.05 if row['dataset'].lower() == 'house' else 0.5,
                                                  random_state=get_default_random_seed())

        schema, test_set_for_predictor = None, test_set
        if 'disc' in row.keys() and bool(row['disc']):
            schema = get_schema(training_set)
            params['discretization'] = schema
            training_set = get_discrete_dataset(training_set.iloc[:, :-1], schema)\
                .join(training_set.iloc[:, -1].reset_index(drop=True))
            test_set_for_predictor = get_discrete_dataset(test_set.iloc[:, :-1], schema) \
                .join(test_set.iloc[:, -1].reset_index(drop=True))

        # Handle Cart tests.
        # Cart needs to inspect the tree of the predictor.
        # Unfortunately onnx does not provide a method to do that.
        if row['predictor'].lower() not in ['dtc', 'dtr']:
            params['predictor'] = Predictor.load_from_onnx(str(get_predictor_path(row['predictor'])))
        else:
            tree = get_model(row['predictor'], {})
            tree.fit(training_set.iloc[:, :-1], training_set.iloc[:, -1])
            params['predictor'] = tree

        # Handle GridEx tests
        # TODO: this is algorithm specific therefore it should be handled inside the algorithm itself.
        if 'grid' in row.keys() and bool:
            strategy, n = eval(row['strategies'])
            if strategy == "F":
                params['grid'] = Grid(int(row['grid']), FixedStrategy(n))
            else:
                ranked = FeatureRanker(training_set.columns[:-1])\
                    .fit(params['predictor'], training_set.iloc[:, :-1]).rankings()
                params['grid'] = Grid(int(row['grid']), AdaptiveStrategy(ranked, n))

        extractor = get_extractor(row['extractor_type'], params)
        mapping = None if 'output_mapping' not in row.keys() or row['output_mapping'] == '' else ast.literal_eval(row['output_mapping'])
        theory = extractor.extract(training_set, mapping) if mapping is not None else extractor.extract(training_set)

        # Compute predictions from rules
        index = test_set.shape[1] - 1
        ordered_test_set = test_set.copy()
        ordered_test_set.iloc[:, :-1] = ordered_test_set.iloc[:, :-1].reindex(sorted(ordered_test_set.columns[:-1]), axis=1)
        is_classification = isinstance(test_set.iloc[0, -1], str)
        cast: Callable = lambda x: (str(x) if is_classification else float(x.value))
        solver = prolog_solver(static_kb=mutable_theory(theory).assertZ(get_in_rule()).assertZ(get_not_in_rule()))
        substitutions = [solver.solveOnce(data_to_struct(data)) for _, data in ordered_test_set.iterrows()]
        expected = [cast(query.solved_query.get_arg_at(index)) for query in substitutions if query.is_yes]
        if mapping is not None:
            predictions = [prediction for prediction in extractor.predict(test_set_for_predictor.iloc[:, :-1], mapping)
                          if prediction is not None]
        else:
            predictions = [prediction for prediction in extractor.predict(test_set_for_predictor.iloc[:, :-1])
                           if prediction is not None]

        yield {
            'extractor': extractor,
            'extracted_theory': theory,
            'extracted_test_y_from_theory': np.array(expected),
            'extracted_test_y_from_extractor': np.array(predictions),
            'test_set': test_set,
            'expected_theory': parse_theory(row['theory'] + '.') if row['theory'] != '' else None,
            'discretization': schema
        }


class Predictor:

    def __init__(self, model, from_file_onnx=False):
        self._model = model
        self._from_file_onnx = from_file_onnx

    @staticmethod
    def load_from_onnx(file: str) -> Predictor:
        return Predictor(rt.InferenceSession(file), True)

    def save_to_onnx(self, file, initial_types: list[tuple[str, DataType]]):
        file = str(file) + '.onnx'
        if not self._from_file_onnx:
            if os.path.exists(file):
                os.remove(file)
            if isinstance(self._model, Model):
                save(self._model, "tmp_model")
                os.system("python -m tf2onnx.convert --saved-model tmp_model --output " + file)
            else:
                onnx_predictor = convert_sklearn(self._model, initial_types=initial_types)
                with open(file, 'wb') as f:
                    f.write(onnx_predictor.SerializeToString())

    def predict(self, dataset: pd.DataFrame | np.ndarray) -> Iterable:
        array = dataset.to_numpy() if isinstance(dataset, pd.DataFrame) else dataset
        if self._from_file_onnx:
            input_name = self._model.get_inputs()[0].name
            label_name = self._model.get_outputs()[0].name
            if array.dtype == 'float64':
                tensor_type = np.float32
            elif array.dtype == 'int64' or array.dtype == 'int32':
                tensor_type = np.int64
            else:
                tensor_type = np.str
            pred_onx = self._model.run([label_name], {input_name: array.astype(tensor_type)})[0]
            return [prediction for plist in pred_onx for prediction in plist] if isinstance(pred_onx[0], list)\
                else [prediction for prediction in pred_onx]
        else:
            return self._model.predict(dataset)

    # TODO: to be improved, make it more flexible
    @staticmethod
    def get_initial_types(dataset: pd.DataFrame | np.ndarray) -> list[tuple[str, DataType]]:
        array = dataset.to_numpy() if isinstance(dataset, pd.DataFrame) else dataset
        name = ''
        for column in dataset.columns:
            name += column + ', '
        name = name[:-2]
        shape = [None, array.shape[1]]
        if array.dtype == 'float64':
            types = FloatTensorType(shape)
        elif array.dtype == 'int64':
            types = Int64TensorType(shape)
        else:
            types = StringTensorType(shape)
        return [(name, types)]



from __future__ import annotations
from typing import Iterable, Callable
import pandas as pd
from tuprolog.core import Var, Struct, Real, Term, Integer, Numeric, clause
import re
from tuprolog.core import struct, real, atom, var, numeric, logic_list, Clause
from tuprolog.core.operators import DEFAULT_OPERATORS, operator, operator_set, XFX
from tuprolog.core.formatters import TermFormatter
from tuprolog.core.visitors import AbstractTermVisitor
from tuprolog.theory import mutable_theory, Theory
from tuprolog.theory.parsing import DEFAULT_CLAUSES_PARSER

from psyke.schema import Value, LessThan, GreaterThan, Between, Constant, term_to_value, Outside
from psyke import DiscreteFeature
from psyke.utils import get_int_precision

LE, GE, L, G = '=<', '>=', '<', '>'

PRECISION: int = get_int_precision()

OP_IN = operator('in', XFX, 700)

OP_NOT = operator('not_in', XFX, 700)

RULES_OPERATORS = DEFAULT_OPERATORS + operator_set(OP_IN, OP_NOT)

REGEX = r'(?<=\.\d\d)\d*(?=(\D|\b))'


def is_sum(term: Struct) -> bool:
    return term.getArity() == 2 and term.getFunctor() == '+'


def is_mult(term: Struct) -> bool:
    return term.getArity() == 2 and term.getFunctor() == '*'


def is_negative(term: Term) -> bool:
    if isinstance(term, Integer):
        return term < Integer.ZERO
    if isinstance(term, Real):
        return term < Real.ZERO
    if isinstance(term, Struct) and is_mult(term):
        return any(map(is_negative, term.getArgs()))
    return False


def is_zero(term: Term) -> bool:
    if isinstance(term, Integer):
        return term == Integer.ZERO
    if isinstance(term, Real):
        return term == Real.ZERO
    if isinstance(term, Struct) and is_mult(term):
        return any(map(is_zero, term.getArgs()))
    return False


def absolute(term: Term) -> bool:
    if is_negative(term):
        if isinstance(term, Numeric):
            return numeric(float(str(term.getValue().unaryMinus())))
        return struct(term.getFunctor(), map(absolute, term.getArgs()))
    return term


class Simplifier(AbstractTermVisitor):
    def defaultValue(self, term):
        return term

    def visitStruct(self, term):
        args = term.getArgs()
        if is_sum(term):
            left, right = args
            if is_mult(right):
                if is_negative(right):
                    return struct('-', map(absolute, args))
            right_left, right_right = right
            if is_sum(right) and is_negative(right_left):
                return struct(
                    '-',
                    left,
                    struct(right.getFunctor(), absolute(right_left), right_right).accept(self)
                )
        return struct(term.getFunctor(), [a.accept(self) for a in args])


def create_functor(constraint: Value, positive: bool = True) -> str:
    if isinstance(constraint, LessThan):
        return '=<' if positive else '>'
    if isinstance(constraint, GreaterThan):
        return '>' if positive else '=<'
    if isinstance(constraint, Between):
        return 'in' if positive else 'not_in'
    if isinstance(constraint, Outside):
        return 'not_in' if positive else 'in'
    if isinstance(constraint, Constant):
        return '=' if positive else '\\='


def _create_term(v: Var, constraint: Value, functor: str) -> Struct:
    if isinstance(constraint, LessThan):
        return struct(functor, v, real(round(constraint.value, PRECISION)))
    if isinstance(constraint, GreaterThan):
        return struct(functor, v, real(round(constraint.value, PRECISION)))
    if isinstance(constraint, Between):
        return struct(functor, v,
                      logic_list(real(round(constraint.lower, PRECISION)), real(round(constraint.upper, PRECISION))))
    if isinstance(constraint, Outside):
        return struct(functor, v,
                      logic_list(real(round(constraint.lower, PRECISION)), real(round(constraint.upper, PRECISION))))
    if isinstance(constraint, Constant):
        return struct(functor, v, atom(str(constraint.value)))


def create_term(v: Var, constraint: Value, positive: bool = True) -> Struct:
    if v is None:
        raise Exception('IllegalArgumentException: None variable')
    functor = create_functor(constraint, positive)
    return _create_term(v, constraint, functor)


def to_var(name: str) -> Var:
    return var(name[0].upper() + name[1:])


def create_variable_list(features: list[DiscreteFeature], dataset: pd.DataFrame = None, sort: bool = True) -> dict[str, Var]:
    if sort:
        features = sorted(features, key=lambda x: x.name)
        dataset = sorted(dataset.columns[:-1]) if dataset is not None else None
    else:
        dataset = dataset.columns[:-1] if dataset is not None else None
    values = {feature.name: to_var(feature.name) for feature in features} \
        if len(features) > 0 else {name: to_var(name) for name in dataset}
    return values


def create_head(functor: str, variables: Iterable[Var], output) -> Struct:
    if isinstance(output, Var):
        variables += [output]
    elif isinstance(output, str):
        variables += [atom(output)]
    else:
        variables += [numeric(round(float(output), PRECISION))]
    return struct(functor, variables)


def pretty_clause(clause: Clause, new_line: bool = True) -> str:
    formatter = TermFormatter.prettyExpressions(True, RULES_OPERATORS)
    if clause.is_fact:
        return str(formatter.format(clause.head))
    elif clause.is_directive:
        return ":- " + formatter.format(clause.body)
    else:
        head = str(formatter.format(clause.head))
        body = str(formatter.format(clause.body))
        return f"{re.sub(REGEX, '', head)} :-\n    {re.sub(REGEX, '', body)}"


def pretty_theory(theory: Theory, new_line: bool = True) -> str:
    if len(str(theory)) == 0:
        return ""
    else:
        clause = [pretty_clause(clause, new_line) for clause in theory]
        return ".\n".join(clause) + "."


def to_rounded_real(n: float) -> Real:
    return real(round(n, PRECISION))


def foldr(accumulator, iterable, default=None):
    items = list(iterable)
    if len(items) == 0:
        return default
    current = items[-1]
    items.pop(-1)
    while len(items) > 0:
        current = accumulator(items[-1], current)
        items.pop(-1)
    return current


def linear_function_creator(features: list[Var], weights: Iterable[Real], intercept: Real) -> Struct:
    x = zip(features[: -1], weights)
    x = filter(lambda fw: not is_zero(fw[1]), x)
    x = map(lambda fw: struct('*', fw[1], fw[0]), x)
    x = foldr(lambda a, b: struct('+', a, b), x)
    x = intercept if x is None else struct('+', intercept, x)
    return struct('is', features[-1], x)


def terms_to_intervals(terms) -> dict[str, Value]:
    return {term.args[0].name: term_to_value(term) for term in terms if term.arity > 0}


def terms_to_minimal_intervals(terms) -> dict[Var, tuple[Value, str]]:
    intervals = [(term.args[0], term_to_value(term), term.functor) for term in terms if term.arity > 0]
    intervals_per_variable: dict[Var: list[tuple[Value, str]]] = {}
    for v, value, functor in intervals:
        if v in intervals_per_variable.keys():
            intervals_per_variable[v].append((value, functor))
        else:
            intervals_per_variable[v] = [(value, functor)]

    result: dict[Var, tuple[Value, str]] = {}
    for v, values in intervals_per_variable.items():
        if len(values) == 1:
            result[v] = values[0]
        else:
            minimal_interval = None
            for value, functor in values:
                minimal_interval = value * minimal_interval
            result[v] = (minimal_interval, create_functor(minimal_interval))
    return result


def is_subset(first_sets: dict[str, Value], second_sets: dict[str, Value]) -> bool:
    def is_inside(first_value: Value, second_value: Value) -> bool:
        return first_value in second_value

    return all(is_inside(v, second_sets[k]) if k in second_sets.keys() else True for k, v in first_sets.items())


def prune(theory: Theory) -> Theory:
    """
    Prune unnecessary clauses from a logic theory T.
    This is a work in progress because it is not a trivial problem.
    Firstly, we remove redundant clauses that are easy to detect.
    Then, we will continue with more complicated ones.

    1. A clause c1 in T is removable when it is followed or preceded by a clause c2 that:
        - has the same head of c1;
        - c2 is always true if c1 is true (c1 -> c2).
    Examples:
        c1(A, B, C, D, positive) :- A =< 1, B > 2, C = 0.
        c2(A, B, C, D, positive) :- A =< 1, B > 2.
        In this case c1 is redundant.
    But also:
        c1(A, B, C, D, positive) :- A =< 1, B > 2, C = 0.
        c2(A, B, C, D, positive) :- A =< 1.3, B > 1.8.
        c1 can be removed.
    And also:
        c2(A, B, C, D, positive) :- A =< 1.3, B > 1.8.
        c1(A, B, C, D, positive) :- A =< 1, B > 2, C = 0.
        c1 can be removed.

    :param theory: the logic theory
    :return: a new simplified theory
    """

    def is_clause_included(clause, other):
        terms_clause = clause.body.unfolded if clause.body.is_recursive else [clause.body]
        terms_other = other.body.unfolded if other.body.is_recursive else [other.body]
        if clause != other and clause.head.args[-1] == other.head.args[-1]:
            set_clause = terms_to_intervals(terms_clause)
            set_other = terms_to_intervals(terms_other)
            return is_subset(set_clause, set_other) and set(set_other.keys()).issubset(set(set_clause.keys()))
        else:
            return False

    def attack(clause, clauses, index):
        after = index < len(clauses) - 1 and clause.body_size > 0 and is_clause_included(clause, clauses[index + 1])
        before = index > 0 and clause.body_size > 0 and is_clause_included(clause, clauses[index - 1])
        return after or before

    new_theory = mutable_theory()
    theory_copy = mutable_theory(theory)
    clauses_copy = theory_copy.clauses
    for i, clause in enumerate(theory.clauses):
        if not attack(clause, clauses_copy, i):
            new_theory.assertZ(clause)
    return new_theory


def simplify(theory: Theory) -> Theory:

    def simplify_clause(c: Clause) -> Clause:

        terms = c.body.unfolded if c.body.is_recursive else [c.body]
        minimal_intervals = terms_to_minimal_intervals(terms)
        minimal_terms = [_create_term(v, value[0], value[1]) for v, value in minimal_intervals.items()]
        return clause(c.head, minimal_terms) if c.body.arity > 0 else c

    new_theory = mutable_theory()
    for old_clause in theory.clauses:
        new_clause = simplify_clause(old_clause)
        new_theory.assertZ(new_clause)
    return new_theory


def data_to_struct(data: pd.Series):
    head = data.keys()[-1]
    terms = [numeric(item) for item in data.values[:-1]]
    terms.append(var('X'))
    return struct(head, terms)


def get_in_rule(min_included: bool = True, max_included: bool = False) -> Clause:
    """
    Create the logic 'in' predicate in(X, [Min, Max]).
    The predicate is true if X is in between Min and Max.
    :param min_included: if X == Min then true
    :param max_included: if X == Max then true
    :return: the tuProlog clause for the 'in' predicate
    """
    in_textual_rule: Callable = lambda x, y: "in(X, [Min, Max]) :- !, X " + x + " Min, X " + y + " Max."
    parser = DEFAULT_CLAUSES_PARSER
    theory = parser.parse_clauses(in_textual_rule(GE if min_included else G, LE if max_included else L), operators=None)
    return theory[0]


def get_not_in_rule(min_included: bool = False, max_included: bool = True) -> Clause:
    """
    Create the logic 'not_in' predicate not_in(X, [Min, Max]).
    The predicate is true if X is outside the range between Min and Max.
    :param min_included: if X == Min then true
    :param max_included: if X == Max then true
    :return: the tuProlog clause for the 'not_in' predicate
    """
    not_in_textual_rule: Callable = lambda x, y: "not_in(X, [Min, Max]) :- X " + x + " Min; X " + y + " Max."
    parser = DEFAULT_CLAUSES_PARSER
    theory = parser.parse_clauses(not_in_textual_rule(LE if min_included else L, GE if max_included else G),
                                  operators=None)
    return theory[0]


import platform
import re

from setuptools import setup, find_packages
import pathlib
import sys
import os
import subprocess
import distutils.cmd

here = pathlib.Path(__file__).parent.resolve()

version_file = here / 'VERSION'

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')


EPOCHS: int = 50
BATCH_SIZE: int = 16
REQUIREMENTS = [
    'numpy~=1.24.0',
    'pandas~=1.5.0',
    'scikit-learn~=1.2.0',
    '2ppy~=0.4.0',
    'kneed~=0.8.1',
    'kivy~=2.1.0',
    'matplotlib~=3.6.0',
    'kivy-garden~=0.1.5',
    'screeninfo~=0.8',
    'sympy~=1.11'
]  # Optional


def format_git_describe_version(version):
    if '-' in version:
        splitted = version.split('-')
        tag = splitted[0]
        index = f"dev{splitted[1]}"
        return f"{tag}.{index}"
    else:
        return version


def get_version_from_git():
    try:
        process = subprocess.run(["git", "describe"], cwd=str(here), check=True, capture_output=True)
        version = process.stdout.decode('utf-8').strip()
        version = format_git_describe_version(version)
        with version_file.open('w') as f:
            f.write(version)
        return version
    except subprocess.CalledProcessError:
        if version_file.exists():
            return version_file.read_text().strip()
        else:
            return '0.1.0.archeo'


version = get_version_from_git()


print(f"Detected version {version} from git describe")


class GetVersionCommand(distutils.cmd.Command):
    """A custom command to get the current project version inferred from git describe."""

    description = 'gets the project version from git describe'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print(version)


class InstallGardenDependency(distutils.cmd.Command):
    """A custom command to install Garden dependencies"""

    garden_executable = pathlib.Path(sys.executable).parent / 'garden'

    description = 'runs `garden install [--app] PACKAGE`'
    user_options = [
        ('package=', None, 'the package to be installed'),
        ('as-library', None, 'wether to install the dependency as a local library')
    ]

    def initialize_options(self):
        self.package = None
        self.as_library = False

    def finalize_options(self):
        if self.package is None:
            raise ValueError("No Garden package provided")

    def run(self):
        cmd = f"{sys.executable} {InstallGardenDependency.garden_executable} install "
        if self.as_library:
            cmd += "--app "
            os.chdir(here / "psyke" / "gui")
        cmd += self.package
        os.system(cmd)


class CreateTestPredictors(distutils.cmd.Command):
    description = 'gets the project version from git describe'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from test.psyke import Predictor
        from psyke.utils import get_default_random_seed
        from psyke.utils.dataframe import get_discrete_dataset
        from sklearn.model_selection import train_test_split
        from test import REQUIRED_PREDICTORS, get_dataset, get_model, get_schema
        from test.resources.predictors import get_predictor_path, PATH, create_predictor_name
        import ast
        import pandas as pd
        from tensorflow.keras import Model

        # Read the required predictors to run the tests:
        #   model | model_options | dataset
        required_predictors = pd.read_csv(REQUIRED_PREDICTORS, sep=';')

        # Create missing predictors.
        #     model | model_options | dataset
        for index, row in required_predictors.iterrows():
            options = ast.literal_eval(row['model_options'])
            file_name = create_predictor_name(row['dataset'], row['model'], options)
            if not get_predictor_path(file_name).is_file():
                dataset = get_dataset(row['dataset'])
                if row['bins'] > 0:
                    schema = get_schema(dataset)  # int(row['bins'])
                    dataset = get_discrete_dataset(dataset.iloc[:, :-1], schema).join(dataset.iloc[:, -1])
                model = get_model(row['model'], options)
                training_set, test_set = train_test_split(dataset, test_size=0.5,
                                                          random_state=get_default_random_seed())
                if isinstance(model, Model):
                    keys = set(training_set.iloc[:, -1])
                    mapping = {key: i for i, key in enumerate(keys)}
                    training_set.iloc[:, -1] = training_set.iloc[:, -1].apply(lambda x: mapping[x])
                    test_set.iloc[:, -1] = test_set.iloc[:, -1].apply(lambda x: mapping[x])
                    model.fit(training_set.iloc[:, :-1], training_set.iloc[:, -1], epochs=EPOCHS, batch_size=BATCH_SIZE)
                else:
                    model.fit(training_set.iloc[:, :-1], training_set.iloc[:, -1])
                predictor = Predictor(model)
                predictor.save_to_onnx(PATH / file_name, Predictor.get_initial_types(training_set.iloc[:, :-1]))

        required_predictors.to_csv(REQUIRED_PREDICTORS, sep=';', index=False)

        print("Done")


class CreateTheoryPlot(distutils.cmd.Command):
    description = 'create a plot representing samples X and their class/regression value Y predicted by a theory'
    user_options = [('theory=', 't', 'textual file of a Prolog theory'),
                    ('dataset=', 'd', 'file of a dataset'),
                    ('azimuth=', 'a', 'azimuth of the plot'),
                    ('distance=', 'D', 'distance from the plot'),
                    ('elevation=', 'e', 'elevation of the plot'),
                    ('output=', 'o', 'output file name of the plot'),
                    ('show=', 's', 'show theory in the plot ([y]/n)'),
                    ]
    default_output_file_name = 'dummy/plot'
    default_theory_name = 'dummy/iris-theory'
    default_dataset_name = 'dummy/iris'
    default_azimuth = '45'
    default_distance = '9'
    default_elevation = '5'
    csv_format = '.csv'
    txt_format = '.txt'
    pdf_format = '.pdf'

    def initialize_options(self):
        self.output = self.default_output_file_name
        self.theory = self.default_theory_name
        self.dataset = self.default_dataset_name
        self.azimuth = self.default_azimuth
        self.elevation = self.default_elevation
        self.distance = self.default_distance
        self.show = True

    def finalize_options(self):
        self.theory_file = str(self.theory)
        self.data = str(self.dataset)
        self.output = str(self.output)
        self.a = float(self.azimuth)
        self.e = float(self.elevation)
        self.d = float(self.distance)
        self.s = self.show in (True, 'y', 'Y', 'yes', 'YES', 'Yes')

    def run(self):
        import pandas as pd
        from tuprolog.theory.parsing import parse_theory
        from psyke.utils.plot import plot_theory

        if self.theory_file is None or self.theory_file == '':
            raise Exception('Empty theory file name')
        if self.data is None or self.data == '':
            raise Exception('Empty dataset file name')
        with open(self.theory_file + (self.txt_format if '.' not in self.theory_file else ''), 'r') as file:
            textual_theory = file.read()
        theory = parse_theory(textual_theory)
        data = pd.read_csv(self.data + (self.csv_format if '.' not in self.data else ''))
        plot_theory(theory, data, self.output + self.pdf_format, self.a, self.d, self.e, self.s)


setup(
    name='psyke',  # Required
    version=version,
    description='Python-based implementation of PSyKE, i.e. a Platform for Symbolic Knowledge Extraction',
    license='Apache 2.0 License',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/psykei/psyke-python',
    author='Matteo Magnini',
    author_email='matteo.magnini@unibo.it',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Prolog'
    ],
    keywords='knowledge extraction, symbolic ai, ske, extractor, rules, prolog',  # Optional
    # package_dir={'': 'src'},  # Optional
    packages=find_packages('.'),  # Required
    include_package_data=True,
    python_requires='>=3.9.0',
    install_requires=REQUIREMENTS,  # Optional
    zip_safe = False,
    platforms = "Independant",
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/psykei/psyke-python/issues',
        # 'Funding': 'https://donate.pypi.org',
        # 'Say Thanks!': 'http://saythanks.io/to/example',
        'Source': 'https://github.com/psykei/psyke-python',
    },
    cmdclass={
        'get_project_version': GetVersionCommand,
        'create_test_predictors': CreateTestPredictors,
        'create_theory_plot': CreateTheoryPlot,
        'garden_install': InstallGardenDependency
    },
)



import numpy as np
import pandas as pd
from matplotlib.lines import Line2D
from matplotlib import pyplot as plt
from matplotlib import colors
from numpy.core.defchararray import capitalize

from psyke.gui.view import MARKERS, COLORS, COLOR_MAPS

plt.rcParams.update({'font.size': 9})

size = (3.75, 3.25)


def init_plot(x, y, title):
    def capitalize_label(label):
        return f'{capitalize(label[0])}{label[1:]}'

    plt.figure(figsize=size)
    plt.title(title)
    plt.gcf().patch.set_alpha(0.0)
    plt.gca().set_xlabel(capitalize_label(x.name))
    plt.gca().set_ylabel(capitalize_label(y.name))
    plt.xlim(x.min(), x.max())
    plt.ylim(y.min(), y.max())
    plt.gca().set_rasterized(True)
    plt.tight_layout()


def plot_regions(x: pd.Series, y: pd.Series, z: pd.Series, colormap):
    idx = [value is not None for value in z]
    if isinstance(z[idx][0], str):
        unique = np.unique(z[idx])
        classes_dict = {c: i for i, c in enumerate(unique)}
        bounds = np.array([i - 0.5 for i in range(len(unique) + 1)])
        norm = colors.BoundaryNorm(bounds, len(unique))
        plt.tricontourf(x[idx], y[idx], [classes_dict[v] for v in z[idx]],
                        cmap=colormap, norm=norm, alpha=.5, levels=bounds)
    else:
        plt.tricontourf(x[idx], y[idx], z[idx], cmap=colormap, alpha=.5, levels=10)


def plotSamples(x: pd.Series, y: pd.Series, z: pd.Series, colormap=None, index=0, name: str = None):
    if isinstance(z[0], str):
        legend_marks = [Line2D([0], [0], color=c, marker=m, lw=0) for c, m in zip(COLORS[index], MARKERS[index])]
        classes = np.unique(z)
        for i, c in enumerate(classes):
            idx = [output == c for output in z]
            plt.scatter(x[idx], y[idx], c=COLORS[index][i], marker=MARKERS[index][i])
        plt.gca().legend(legend_marks, classes)
    else:
        sc = plt.scatter(x, y, c=z, cmap=colormap,
                         edgecolor="k", linewidths=0.05, s=7)
        plt.colorbar(sc, pad=0.03, label=z.name)
    if name is not None:
        plt.savefig(name, dpi=500, bbox_inches='tight')


def create_grid(x_name: str, y_name: str, data: pd.DataFrame):
    n = [100 if feature in [x_name, y_name] else 10 for feature in data.columns[:-1]]
    vectors = [np.linspace(data[feature].min(), data[feature].max(), nn) if feature in [x_name, y_name] else
               data[feature].mean() for feature, nn in zip(data.columns[:-1], n)]
    mesh = np.meshgrid(*vectors)
    df = pd.DataFrame(np.c_[[vector.ravel() for vector in mesh]]).T
    df.columns = data.columns[:-1]
    return df



import unittest
from parameterized import parameterized_class
from psyke import logger
from test.psyke import initialize


@parameterized_class(initialize('real'))
class TestReal(unittest.TestCase):

    def test_extract(self):
        logger.info(self.expected_theory)
        logger.info(self.extracted_theory)
        self.assertTrue(self.expected_theory.equals(self.extracted_theory, False))

    def test_predict(self):
        self.assertTrue(all(self.extracted_test_y_from_theory == self.extracted_test_y_from_extractor))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
import math
from typing import Callable
from psyke.utils import get_int_precision

_EMPTY_INTERSECTION_EXCEPTION: Callable = lambda x, y: \
    Exception("Empty intersection between two Value: " + str(x) + ' and ' + str(y))

_NOT_IMPLEMENTED_INTERSECTION: Callable = lambda x, y: \
    Exception("Not implemented intersection between: " + str(x) + ' and ' + str(y))

_INTERSECTION_WITH_WRONG_TYPE: Callable = lambda x, y: \
    Exception("Calling method with wrong type argument: " + str(x) + ' and ' + str(y))

PRECISION = get_int_precision()
STRING_PRECISION = str(PRECISION)


class DiscreteFeature:

    def __init__(self, name: str, admissible_values: dict[str, Value]):
        self.name = name
        self.admissible_values = admissible_values

    def __str__(self):
        return self.name + " = {" \
            + ", ".join([f"'{k}' if {self.name} ∈ {str(self.admissible_values[k])}" for k in self.admissible_values]) \
            + "}"

    def __repr__(self):
        return f"DiscreteFeature(name={self.name}, admissible_values={self.admissible_values})"


class Value:

    def __init__(self):
        pass

    def is_in(self, other: float) -> bool:
        """
        Check if a real number is inside an interval, or it is equal to a constant.

        :param other: the value to check
        :return: true if the value is inside the interval, false otherwise
        """
        return False

    def is_boundary(self, other: float) -> bool:
        """
        Check if a real number is one edge of an interval, or it is equal to a constant.

        :param other: the value to check
        :return: true if the value is one edge the interval, false otherwise
        """
        if isinstance(self, Constant):
            return self.value == other
        elif isinstance(self, Interval):
            return self.lower == other or self.upper == other
        else:
            return False

    def is_in_or_is_boundary(self, other_value: float) -> bool:
        """
        Check if a real number is in or is boundary for an interval, or for a constant.

        :param other_value: the value to check
        :return: true if at least one condition is true, false otherwise
        """
        return self.is_in(other_value) or self.is_boundary(other_value)

    def __contains__(self, other: Value) -> bool:
        """
        Check if a Value (interval or constant) is contained inside another Value.

        :param other: the other Value
        :return: true if the interval or constant is inside the other interval or constant.
        """
        if isinstance(other, Constant):
            return self.is_in(other.value)
        elif isinstance(other, Interval):
            if type(self) is type(other):
                return self.is_in_or_is_boundary(other.lower) and self.is_in_or_is_boundary(other.upper)
            else:
                return self.is_in(other.lower) and self.is_in(other.upper)
        else:
            return False

    # TODO: handle convention (low priority).
    def __mul__(self, other) -> Value:

        def intersection_with_constant(first_value: Constant, second_value: Value) -> Value:
            if isinstance(first_value, Constant):
                if second_value.is_in(first_value.value):
                    return first_value
                else:
                    raise _EMPTY_INTERSECTION_EXCEPTION(first_value, second_value)
            else:
                raise _INTERSECTION_WITH_WRONG_TYPE(first_value, second_value)

        def intersection_with_outside(first_value: Outside, second_value: Value) -> Value:
            if isinstance(first_value, Outside):
                if isinstance(second_value, LessThan):
                    if second_value.value <= first_value.lower:
                        return second_value
                    elif first_value.is_in(second_value.value):
                        return LessThan(first_value.lower)
                    else:
                        raise _NOT_IMPLEMENTED_INTERSECTION(first_value, second_value)
                elif isinstance(second_value, GreaterThan):
                    if second_value.value >= first_value.lower:
                        return GreaterThan(first_value.upper)
                    elif first_value.is_in(second_value.value):
                        return second_value
                    else:
                        raise _NOT_IMPLEMENTED_INTERSECTION(first_value, second_value)
                elif isinstance(second_value, Constant):
                    if not first_value.is_in(second_value.value):
                        return second_value
                    else:
                        raise _EMPTY_INTERSECTION_EXCEPTION(first_value, second_value)
                elif isinstance(second_value, Between):
                    if second_value in first_value:
                        return second_value
                    elif second_value.lower <= first_value.lower <= second_value.upper <= first_value.upper:
                        return Between(second_value.lower, first_value.lower)
                    elif first_value.lower <= second_value.lower <= first_value.upper <= second_value.upper:
                        return Between(first_value.upper, second_value.upper)
                    else:
                        raise _NOT_IMPLEMENTED_INTERSECTION(first_value, second_value)
                elif isinstance(second_value, Outside):
                    if second_value.lower <= first_value.lower and second_value.upper >= first_value.upper:
                        return second_value
                    elif first_value.lower <= second_value.lower and first_value.upper >= second_value.upper:
                        return first_value
                    else:
                        raise _EMPTY_INTERSECTION_EXCEPTION(first_value, second_value)
                elif isinstance(second_value, Constant):
                    intersection_with_constant(second_value, first_value)
                else:
                    raise _INTERSECTION_WITH_WRONG_TYPE(first_value, second_value)
            else:
                raise _INTERSECTION_WITH_WRONG_TYPE(first_value, second_value)

        def intersection_with_between(first_value: Between, second_value: Value) -> Value:
            if isinstance(first_value, Between):
                if isinstance(second_value, LessThan):
                    if second_value.value <= first_value.lower:
                        raise _EMPTY_INTERSECTION_EXCEPTION(first_value, second_value)
                    elif first_value.lower <= second_value.value <= first_value.upper:
                        return Between(first_value.lower, second_value.value)
                    else:
                        return first_value
                elif isinstance(second_value, GreaterThan):
                    if second_value.value <= first_value.lower:
                        return first_value
                    elif first_value.lower <= second_value.value <= first_value.upper:
                        return Between(second_value.value, first_value.upper)
                    else:
                        raise _EMPTY_INTERSECTION_EXCEPTION(first_value, second_value)
                elif isinstance(second_value, Between):
                    if second_value in first_value:
                        return second_value
                    elif first_value in second_value:
                        return first_value
                    elif first_value.lower <= second_value.lower <= first_value.upper:
                        return Between(second_value.lower, first_value.upper)
                    elif second_value.lower <= first_value.lower <= second_value.upper:
                        return Between(first_value.lower, second_value.upper)
                    elif first_value.lower <= second_value.upper <= first_value.upper:
                        return Between(first_value.lower, second_value.upper)
                    elif second_value.lower <= first_value.upper <= second_value.upper:
                        return Between(second_value.lower, first_value.upper)
                    else:
                        raise _EMPTY_INTERSECTION_EXCEPTION(first_value, second_value)
                elif isinstance(second_value, Constant):
                    intersection_with_constant(second_value, first_value)
                elif isinstance(second_value, Outside):
                    return intersection_with_outside(second_value, first_value)
                else:
                    raise _INTERSECTION_WITH_WRONG_TYPE(first_value, second_value)
            else:
                raise _INTERSECTION_WITH_WRONG_TYPE(first_value, second_value)

        def intersection_with_less_than(first_value: LessThan, second_value: Value) -> Value:
            if isinstance(first_value, LessThan):
                if isinstance(second_value, LessThan):
                    return first_value if first_value in second_value else second_value
                elif isinstance(second_value, GreaterThan):
                    if second_value.value <= first_value.value:
                        return Between(second_value.value, first_value.value)
                    else:
                        raise _EMPTY_INTERSECTION_EXCEPTION(first_value, second_value)
                elif isinstance(second_value, Constant):
                    return intersection_with_constant(second_value, first_value)
                elif isinstance(second_value, Outside):
                    return intersection_with_outside(second_value, first_value)
                elif isinstance(second_value, Between):
                    return intersection_with_between(second_value, first_value)
                else:
                    raise _INTERSECTION_WITH_WRONG_TYPE(first_value, second_value)
            else:
                raise _INTERSECTION_WITH_WRONG_TYPE(first_value, second_value)

        def intersection_with_greater_than(first_value: GreaterThan, second_value: Value) -> Value:
            if isinstance(first_value, GreaterThan):
                if isinstance(second_value, GreaterThan):
                    return first_value if first_value in second_value else second_value
                elif isinstance(second_value, Constant):
                    return intersection_with_constant(second_value, first_value)
                elif isinstance(second_value, Outside):
                    return intersection_with_outside(second_value, first_value)
                elif isinstance(second_value, Between):
                    return intersection_with_between(second_value, first_value)
                elif isinstance(second_value, LessThan):
                    return intersection_with_less_than(second_value, first_value)
                else:
                    raise _INTERSECTION_WITH_WRONG_TYPE(first_value, second_value)
            else:
                raise _INTERSECTION_WITH_WRONG_TYPE(first_value, second_value)

        if other is None:
            return self
        elif isinstance(self, Constant):
            return intersection_with_constant(self, other)
        elif isinstance(self, Outside):
            return intersection_with_outside(self, other)
        elif isinstance(self, Between):
            return intersection_with_between(self, other)
        elif isinstance(self, LessThan):
            return intersection_with_less_than(self, other)
        elif isinstance(self, GreaterThan):
            return intersection_with_greater_than(self, other)
        else:
            raise _INTERSECTION_WITH_WRONG_TYPE(self, other)


class Interval(Value):

    def __init__(self, lower: float, upper: float, standard: bool = True):
        super().__init__()
        self.standard = standard
        self.lower = round(lower, PRECISION)
        self.upper = round(upper, PRECISION)

    def __str__(self):
        return f"[{self.lower:.2f}, {self.upper:.2f}]"

    def __repr__(self):
        return f"Interval({self.lower:.2f}, {self.upper:.2f})"

    def __eq__(self, other: Between) -> bool:
        return (self.upper == other.upper) and (self.lower == other.lower) and (self.standard == other.standard)


class LessThan(Interval):

    def __init__(self, value: float, standard: bool = True):
        super().__init__(-math.inf, value, standard)

    def is_in(self, other: float) -> bool:
        return other <= self.upper if self.standard else other < self.upper

    @property
    def value(self) -> float:
        return self.upper

    def __str__(self):
        return f"]-∞, {self.upper:.2f}" + ("]" if self.standard else "[")

    def __repr__(self):
        return f"LessThan({self.upper:.2f})"

    def __eq__(self, other: LessThan) -> bool:
        return (self.upper == other.upper) and (self.value == other.value) and (self.standard == other.standard)


class GreaterThan(Interval):

    def __init__(self, value: float, standard: bool = True):
        super().__init__(value, math.inf, standard)

    def is_in(self, other: float) -> bool:
        return other > self.lower if self.standard else other >= self.lower

    @property
    def value(self) -> float:
        return self.lower

    def __str__(self):
        return ("]" if self.standard else "[") + f"{self.lower:.2f}, ∞["

    def __repr__(self):
        return f"GreaterThan({self.lower:.2f})"

    def __eq__(self, other: GreaterThan) -> bool:
        return (self.lower == other.lower) and (self.value == other.value) and (self.standard == other.standard)


class Between(Interval):

    def __init__(self, lowerbound: float, upperbound: float, standard: bool = True):
        super().__init__(lowerbound, upperbound, standard)

    def is_in(self, other: float) -> bool:
        return self.lower <= other < self.upper if self.standard else self.lower < other <= self.upper

    def __str__(self):
        return ("[" if self.standard else "]") + f"{self.lower:.2f}, {self.upper:.2f}" + ("[" if self.standard else "]")

    def __repr__(self):
        return f"Between({self.lower:.2f}, {self.upper:.2f})"


class Outside(Interval):

    def __init__(self, lowerbound: float, upperbound: float, standard: bool = True):
        super().__init__(lowerbound, upperbound, standard)

    def is_in(self, other: float) -> bool:
        return other < self.lower or self.upper <= other if self.standard else other <= self.lower or self.upper < other

    def __str__(self):
        return f"]-∞, {self.lower:.2f}" + ("[" if self.standard else "]") + ' U '\
               + ("[" if self.standard else "]") + f"{self.upper:.2f}, ∞["

    def __repr__(self):
        return f"Outside({self.lower:.2f}, {self.upper:.2f})"


class Constant(Value):

    def __init__(self, value: float):
        super().__init__()
        self.value = round(value, get_int_precision())

    def is_in(self, other: float) -> bool:
        return math.isclose(other, self.value)

    def __str__(self):
        return "{" + str(self.value) + "}"

    def __repr__(self):
        return f"Constant({self.value})"


def term_to_value(term) -> Value:

    real_to_float: Callable = lambda x: float(str(x)) if not hasattr(x, 'len') else float(str(x[0]))

    functor_to_value = {
        '<': lambda x: LessThan(real_to_float(x), standard=False),
        '=<': lambda x: LessThan(real_to_float(x)),
        '>': lambda x: GreaterThan(real_to_float(x)),
        '>=': lambda x: GreaterThan(real_to_float(x), standard=False),
        '==': lambda x: Constant(real_to_float(x)),
        'in': lambda x: Between(real_to_float(x[0]), real_to_float(x[1][0])),
        'not_in': lambda x: Outside(real_to_float(x[0]), real_to_float(x[1][0]))
    }
    return functor_to_value[term.functor](term.args[1])



import math
import warnings

warnings.simplefilter("ignore")

Dimension = tuple[float, float]
Dimensions = dict[str, Dimension]


class Expansion:

    def __init__(self, cube, feature: str, direction: str, distance: float = math.nan):
        self.cube = cube
        self.feature = feature
        self.direction = direction
        self.distance = distance

    def __getitem__(self, index: int) -> float:
        return self.cube[self.feature][index]

    def boundaries(self, a: float, b: float) -> (float, float):
        return (self[0], b) if self.direction == '-' else (a, self[1])


class Limit:

    def __init__(self, feature: str, direction: str):
        self.feature = feature
        self.direction = direction

    def __eq__(self, other):
        return (self.feature == other.feature) and (self.direction == other.direction)

    def __hash__(self):
        return hash(self.feature + self.direction)


class MinUpdate:

    def __init__(self, name: str, value: float):
        self.name = name
        self.value = value


class ZippedDimension:

    def __init__(self, name: str, this_dimension: Dimension, other_dimension: Dimension):
        self.name = name
        self.this_dimension = this_dimension
        self.other_dimension = other_dimension



from typing import Union

import pandas as pd
from tuprolog.theory import Theory

from psyke.extraction.hypercubic.hypercube import ClosedRegressionCube, ClosedClassificationCube, ClosedCube
from psyke.utils import Target


class InterpretableClustering:

    def __init__(self, depth: int, error_threshold: float, output: Target = Target.CONSTANT, gauss_components: int = 2):
        self.depth = depth
        self.error_threshold = error_threshold
        self.gauss_components = gauss_components
        self._output = output
        self._hypercubes = []

    def extract(self, dataframe: pd.DataFrame) -> Theory:
        raise NotImplementedError('extract')

    def _default_cube(self) -> Union[ClosedCube, ClosedRegressionCube, ClosedClassificationCube]:
        if self._output == Target.CONSTANT:
            return ClosedCube()
        if self._output == Target.REGRESSION:
            return ClosedRegressionCube()
        return ClosedClassificationCube()



import unittest
from tuprolog.theory import mutable_theory, theory
from tuprolog.theory.parsing import parse_theory
from psyke.utils.logic import simplify


class TestSimplify(unittest.TestCase):

    def test_simplify(self):
        # TODO: if numbers are not float equals method return false (e.g., 2 instead of 2.0). @Giovanni 2ppy
        textual_theory = "p(X, Y, inside) :- ('=<'(X, 1.0), '>'(Y, 2.0), '=<'(X, 0.5))."
        textual_simplified_theory = "p(X, Y, inside) :- ('=<'(X, 0.5), '>'(Y, 2.0))."
        long_theory = mutable_theory(parse_theory(textual_theory))
        simplified_theory = theory(parse_theory(textual_simplified_theory))

        self.assertTrue(simplified_theory.equals(simplify(long_theory), False))


if __name__ == '__main__':
    unittest.main()
